@@ -46,6 +46,7 @@ const DisplayGrid = class {
         this._desktopIndex = desktopIndex;
         this._asDesktop = asDesktop;
         this._desktopDescription = desktopDescription;
+        this._hidden = hidden;
         this._using_X11 = this.DesktopIconsUtil.usingX11();
         this.directoryOpenTimer = null;
         this.windowGlobalRectangle = new Gdk.Rectangle();
@@ -54,7 +55,7 @@ const DisplayGrid = class {
         this._createGrids();
 
         this._window =
-            new Adw.ApplicationWindow(
+            new Gtk.ApplicationWindow(
                 {
                     application: desktopManager.mainApp,
                     'title': desktopName,
@@ -69,31 +70,41 @@ const DisplayGrid = class {
         if (this._asDesktop) {
             this._window.set_decorated(false);
             this._window.set_deletable(false);
+            this._window.set_resizable(false);
 
             // Transparent Background only if this is working as a desktop
             this._window.set_name('desktopwindow');
 
+            this._window
+                .set_default_size(this._windowWidth, this._windowHeight);
+
+            this._window
+                .set_size_request(this._windowWidth, this._windowHeight);
+
+            this._mappedPromise =
+                new Promise(resolve => (this._resolveMapped = resolve));
+
             if (!this._using_X11) {
                 // Wayland Compositer hang on some high resolution
                 // requires all windows be maximized to map and display
                 // initially.
                 this._window.maximize();
+            }
 
-                // However this creates an error where the window can
+            this._window.connect('map', () => {
+                if (!this._resolveMapped)
+                    return;
+                this._resolveMapped(true);
+                this._resolveMapped = null;
+                // Maximize however creates an error where the window can
                 // be moved by the user by dragging down on top panel.
                 // So we unmaximize all windows after they are mapped
-                //  as maximization is not needed anymore.
-                this._window.connect('map', () => this._window.unmaximize());
-            }
+                // as maximization is not needed anymore.
+                this._window.unmaximize();
+            });
         } else {
             // Opaque black test window
             this._window.set_name('testwindow');
-            const headerBar = Adw.HeaderBar.new();
-            const headerTitle = Adw.WindowTitle.new('DING Test Window', '');
-            headerBar.set_title_widget(headerTitle);
-            headerBar.set_show_end_title_buttons(true);
-            this.testbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 0);
-            this.testbox.append(headerBar);
         }
 
         // Remove any other css classes, even if applied by other apps later
@@ -102,8 +113,6 @@ const DisplayGrid = class {
             this._window.set_css_classes(['background']);
         });
 
-        this._window.set_resizable(false);
-
         this._window.connect(
             'close-request',
             () => {
@@ -125,40 +134,89 @@ const DisplayGrid = class {
 
         // New: one fixed root that contains both layers
         this._rootFixed = new Gtk.Fixed();
+        this._rootFixed.set_size_request(this._windowWidth, this._windowHeight);
 
         this._container = new Gtk.Fixed();
         this._containerContext = this._container.get_style_context();
+        this._container.set_size_request(this._windowWidth, this._windowHeight);
         this._containerContext.add_class('unhighlightdroptarget');
-        this._sizeContainer(this._container);
 
         // icon grid goes in rootFixed
         this._rootFixed.put(this._container, 0, 0);
 
         this._overlay = new Gtk.Overlay();
+        this._overlay.set_hexpand(true);
+        this._overlay.set_vexpand(true);
         this._overlay.set_child(this._rootFixed);
-        if (this._asDesktop) {
-            this._window.set_content(this._overlay);
-        } else {
-            this.testbox.append(this._overlay);
-            this._window.set_content(this.testbox);
-        }
+
+        this._window.set_child(this._overlay);
 
         this.gridGlobalRectangle = new Gdk.Rectangle();
-
         this._selectedList = null;
-
         this._setGridStatus();
 
-        if (!hidden)
-            this._window.show();
-        else
-            this._window.hide();
-
-        this._window.set_size_request(this._windowWidth, this._windowHeight);
-
         this._updateGridRectangle();
     }
 
+    ensureMapped() {
+        // show/present only here after the window is fully set up to
+        // and to avoit commiting content too early so that the shell
+        // errors on commiting first frame before acknowleding ack from wayland
+        // compositor.
+        this._window.set_visible(!this._hidden);
+        this._window.present();
+        return this._mappedPromise;
+    }
+
+    ensureAllocationComplete() {
+        if (this._allocPromise)
+            return this._allocPromise;
+
+        const w = this._container;
+
+        this._allocPromise = new Promise(resolve => {
+            let tickId = 0;
+            let stableFrames = 0;
+
+            const cleanup = () => {
+                if (tickId)
+                    w.remove_tick_callback(tickId);
+                this._allocPromise = null;
+            };
+
+            const isAllocated = () => {
+                const aw = w.get_allocated_width();
+                const ah = w.get_allocated_height();
+                return aw > 0 && ah > 0;
+            };
+
+            if (isAllocated()) {
+                this._overlay.queue_draw();
+                resolve();
+                cleanup();
+                return;
+            }
+
+            tickId = w.add_tick_callback(() => {
+                if (isAllocated())
+                    stableFrames++;
+                else
+                    stableFrames = 0;
+
+                if (stableFrames >= 2) {
+                    cleanup();
+                    this._overlay.queue_draw();
+                    resolve();
+                    return GLib.SOURCE_REMOVE;
+                }
+
+                return GLib.SOURCE_CONTINUE;
+            });
+        });
+
+        return this._allocPromise;
+    }
+
     setErrorState() {
         this._window.set_name('errorstate');
     }
@@ -180,6 +238,12 @@ const DisplayGrid = class {
         this._hidden = false;
     }
 
+    queue_draw() {
+        this._container.queue_draw();
+        this._overlay.queue_draw();
+        this._window.queue_draw();
+    }
+
     // Establish and update window geometry, establish and update
     // grid for the desktop icons
 
@@ -218,8 +282,6 @@ const DisplayGrid = class {
         this._window.set_default_size(this._windowWidth, this._windowHeight);
         this._window.set_size_request(this._windowWidth, this._windowHeight);
         this.scale = this._window.get_scale_factor();
-        this._drawArea.set_content_height(this._windowHeight);
-        this._drawArea.set_content_width(this._windowWidth);
     }
 
     _updateUnscaledHeightWidthMargins() {
@@ -330,7 +392,6 @@ const DisplayGrid = class {
         this._container.set_size_request(this._windowWidth, this._windowHeight);
         this._rootFixed.set_size_request(this._windowWidth, this._windowHeight);
         this._sizeContainer(this._container);
-        this._sizeContainer(this._drawArea);
 
         this._updateGridRectangle();
         this._setGridStatus();
@@ -1095,52 +1156,87 @@ const DisplayGrid = class {
     }
 };
 
+const GridOverlay = GObject.registerClass(
+class GridOverlay extends Gtk.Widget {
+    constructor(grid) {
+        super({can_target: false});
+        this._grid = grid;
+    }
+
+    vfunc_snapshot(snapshot) {
+        const w = this.get_allocated_width();
+        const h = this.get_allocated_height();
+
+        if (w <= 0 || h <= 0)
+            return;
+
+        this._grid._doDrawOnGrid(snapshot);
+
+        // Ensure this widget contributes a node so the frame overwrites stale content.
+        const rect = new Graphene.Rect();
+        rect.init(0, 0, 1, 1);
+        snapshot.append_color(
+            new Gdk.RGBA({red: 0, green: 0, blue: 0, alpha: 0.001}),
+            rect
+        );
+    }
+});
+
 const DrawGrid =  class extends DisplayGrid {
     constructor(params) {
         super(params);
 
-        this._drawArea = new Gtk.DrawingArea();
-        this._drawArea.set_content_height(this._windowHeight);
-        this._drawArea.set_content_width(this._windowWidth);
+        this._drawArea = new GridOverlay(this);
+        this._drawArea.set_size_request(this._windowWidth, this._windowHeight);
         this._sizeContainer(this._drawArea);
-        this._drawArea.set_draw_func(this._doDrawOnGrid.bind(this));
         this._overlay.add_overlay(this._drawArea);
         this._drawArea.set_can_target(false);
+        this._drawArea.set_visible(true);
+    }
+
+    resizeWindow() {
+        super.resizeWindow();
+        this._drawArea.set_size_request(this._windowWidth, this._windowHeight);
+    }
+
+    resizeGrid() {
+        super.resizeGrid();
+        this._drawArea.set_size_request(this._windowWidth, this._windowHeight);
+        this._sizeContainer(this._drawArea);
     }
 
     // Functions for drawing on the grid
 
-    queue_draw() {
-        this._drawArea.queue_draw();
-    }
-
     highLightGridAt(x, y) {
         const globalCoordinates = false;
         const selected = this.getCoordinatesOfGridContaining(x, y, globalCoordinates);
         this._selectedList = [selected];
-        this._drawDropRectangles();
+        this.updateOverlay();
     }
 
     unHighLightGrids() {
         this._selectedList = null;
-        this._drawDropRectangles();
+        this.updateOverlay();
     }
 
-    drawRubberBand() {
+    updateOverlay() {
         this._drawArea.queue_draw();
     }
 
-    _doDrawOnGrid(actor, cr) {
-        this._doDrawRubberBand(cr);
-        this._doDrawDropRectangles(cr).catch(console.error);
-        cr.$dispose();
+    _overlayHasContent() {
+        const hasRubberBand =
+            this._dragManager.rubberBand &&
+            this._dragManager.selectionRectangle;
+        const hasDropRects = (this._selectedList?.length ?? 0) > 0;
+        return hasRubberBand || hasDropRects;
     }
 
-    _drawDropRectangles() {
-        this._drawArea.queue_draw();
+    _doDrawOnGrid(snapshot) {
+        this._doDrawRubberBand(snapshot);
+        this._doDrawDropRectangles(snapshot);
     }
 
-    _doDrawRubberBand(cr) {
+    _doDrawRubberBand(snapshot) {
         if (!this._dragManager.rubberBand ||
             !this._dragManager.selectionRectangle ||
             !this.gridGlobalRectangle
@@ -1182,13 +1278,13 @@ const DrawGrid =  class extends DisplayGrid {
             yInit,
             width,
             height,
-            cr,
+            snapshot,
             fillColor,
             outlineColor
         );
     }
 
-    async _doDrawDropRectangles(cr) {
+    _doDrawDropRectangles(snapshot) {
         if (!this.Prefs.showDropPlace || this._selectedList === null)
             return;
 
@@ -1206,90 +1302,71 @@ const DrawGrid =  class extends DisplayGrid {
             alpha: 1.0,
         });
 
-        const dropRectanglePromises =
-            this._selectedList.map(
-                ([x, y]) => {
-                    return this._rectangleDraw(
-                        x, y,
-                        this._elementWidth,
-                        this._elementHeight,
-                        cr,
-                        fillColor,
-                        outlineColor
-                    );
-                }
+        for (const [x, y] of this._selectedList) {
+            this._rectangleDraw(
+                x, y,
+                this._elementWidth,
+                this._elementHeight,
+                snapshot,
+                fillColor,
+                outlineColor
             );
-
-        await Promise.all(dropRectanglePromises).catch(console.error);
+        }
     }
 
-    _rectangleDraw(x, y, width, height, cr, fillColor, outlineColor) {
-        return new Promise(resolve => {
-            cr.rectangle(x + 0.5, y + 0.5, width, height);
-            Gdk.cairo_set_source_rgba(cr, fillColor);
-            cr.fillPreserve();
-            cr.setLineWidth(0.5);
-            Gdk.cairo_set_source_rgba(cr, outlineColor);
-            cr.stroke();
-            resolve(true);
-        });
+    _rectangleDraw(x, y, width, height, snapshot, fillColor, outlineColor) {
+        const rect = new Graphene.Rect();
+        rect.init(x + 0.5, y + 0.5, width, height);
+
+        snapshot.append_color(fillColor, rect);
+
+        const rr = new Gsk.RoundedRect();
+        const zero = new Graphene.Size();
+        zero.init(0, 0);
+        rr.init(rect, zero, zero, zero, zero);
+
+        snapshot.append_border(
+            rr,
+            [0.5, 0.5, 0.5, 0.5],
+            [outlineColor, outlineColor, outlineColor, outlineColor]
+        );
     }
 
-    _roundedRectangleDraw(x, y, width, height, cr, fillColor, outlineColor) {
+    _roundedRectangleDraw(x, y, width, height, snapshot, fillColor, outlineColor) {
         const cornerRadius = 5;
-        const degrees = 3.14 / 180;
 
         const isSquare = width === height;
         const tooLarge = cornerRadius * 2 > Math.min(width, height);
 
-        cr.newSubPath();
+        const useSquareCorners = cornerRadius <= 0 || isSquare || tooLarge;
 
-        if (cornerRadius <= 0 || tooLarge || isSquare) {
-            // Just draw a plain rectangle
-            cr.rectangle(x, y, width, height);
+        const radius =
+            useSquareCorners
+                ? 0
+                : Math.min(cornerRadius, width / 2, height / 2);
+
+        const rect = new Graphene.Rect();
+        rect.init(x, y, width, height);
+
+        const size = new Graphene.Size();
+        size.init(radius, radius);
+
+        const rr = new Gsk.RoundedRect();
+        rr.init(rect, size, size, size, size);
+
+        if (radius > 0) {
+            snapshot.push_rounded_clip(rr);
+            snapshot.append_color(fillColor, rect);
+            snapshot.pop();
         } else {
-            const radius = Math.min(cornerRadius, width / 2, height / 2);
-
-            cr.arc(
-                x + width - radius,
-                y + radius, radius,
-                -90 * degrees,
-                0 * degrees
-            );
-
-            cr.arc(
-                x + width - radius,
-                y + height - radius,
-                radius, 0 * degrees,
-                90 * degrees
-            );
-
-            cr.arc(
-                x + radius,
-                y + height - radius,
-                radius,
-                90 * degrees,
-                180 * degrees
-            );
-
-            cr.arc(
-                x + radius,
-                y + radius,
-                radius,
-                180 * degrees,
-                270 * degrees
-            );
-
-            cr.closePath();
+            snapshot.append_color(fillColor, rect);
         }
 
-        Gdk.cairo_set_source_rgba(cr, fillColor);
-        cr.fillPreserve();
-
-        cr.setLineWidth(1.0);
-        Gdk.cairo_set_source_rgba(cr, outlineColor);
-
-        cr.stroke();
+        snapshot.append_border(
+            rr,
+            [1.0, 1.0, 1.0, 1.0],
+            [outlineColor, outlineColor, outlineColor, outlineColor]
+        );
     }
 };
 
@@ -1301,14 +1378,31 @@ const ControlGrid = class extends DrawGrid {
     }
 
     _addDragControllers() {
+        // Bubble-phase controller: delivers key events to DesktopManager for actions
         this._eventKey = Gtk.EventControllerKey.new();
+        this._eventKey.set_propagation_phase(Gtk.PropagationPhase.BUBBLE);
         this._window.add_controller(this._eventKey);
 
+        // Capture-phase controller: only caches modifier state, does not invoke actions
+        this._eventKeyState = Gtk.EventControllerKey.new();
+        this._eventKeyState.set_propagation_phase(Gtk.PropagationPhase.CAPTURE);
+        this._window.add_controller(this._eventKeyState);
+
         this._eventKey.connect(
             'key-pressed',
             this._onKeyPress.bind(this)
         );
 
+        this._eventKeyState.connect(
+            'key-pressed',
+            this._onModifierUpdate.bind(this)
+        );
+
+        this._eventKeyState.connect(
+            'key-released',
+            this._onModifierClear.bind(this)
+        );
+
         this._eventMotion = Gtk.EventControllerMotion.new();
         this._eventMotion.set_propagation_phase(Gtk.PropagationPhase.BUBBLE);
         this._container.add_controller(this._eventMotion);
@@ -1325,12 +1419,10 @@ const ControlGrid = class extends DrawGrid {
             }
         );
 
-        this._buttonClick = Gtk.GestureClick.new();
-        this._buttonClick.set_button(0);
+        this._buttonClick = new Gtk.GestureClick({button: 0});
         this._buttonClick.set_propagation_phase(Gtk.PropagationPhase.BUBBLE);
         this._container.add_controller(this._buttonClick);
-        this._buttonLongClick = Gtk.GestureLongPress.new();
-        this._buttonLongClick.set_button(0);
+        this._buttonLongClick = new Gtk.GestureLongPress({button: 0});
         this._buttonLongClick.set_propagation_phase(Gtk.PropagationPhase.BUBBLE);
         this._container.add_controller(this._buttonLongClick);
 
@@ -1372,6 +1464,14 @@ const ControlGrid = class extends DrawGrid {
         );
     }
 
+    _onModifierUpdate(_actor, _keyval, _keycode, state) {
+        this._desktopManager.updateModifierState(state);
+    }
+
+    _onModifierClear() {
+        this._desktopManager.clearModifierState();
+    }
+
     _doGesturePress(actor, nPress, x, y) {
         if (this._desktopManager.closePopUps())
             return;
@@ -2104,7 +2204,7 @@ const ControlGrid = class extends DrawGrid {
 
         if (selectedList === null) {
             this._selectedList = null;
-            this._drawDropRectangles();
+            this.updateOverlay();
 
             return;
         }
@@ -2131,7 +2231,7 @@ const ControlGrid = class extends DrawGrid {
         if (newSelectedList.length === 0) {
             if (this._selectedList !== null) {
                 this._selectedList = null;
-                this._drawDropRectangles();
+                this.updateOverlay();
             }
 
             return;
@@ -2145,7 +2245,7 @@ const ControlGrid = class extends DrawGrid {
         }
 
         this._selectedList = newSelectedList;
-        this._drawDropRectangles();
+        this.updateOverlay();
     }
 
     _startSpringLoadedTimer(fileItem) {
@@ -2341,13 +2441,11 @@ const WidgetGrid = class extends ControlGrid {
         this._widgetContainer.add_controller(drag);
 
         // Click gesture: used only to track selection + click radius
-        const click = new Gtk.GestureClick();
-        click.set_button(0);
+        const click = new Gtk.GestureClick({button: 0});
         click.set_propagation_phase(Gtk.PropagationPhase.CAPTURE);
         this._widgetContainer.add_controller(click);
 
-        const contextClick = new Gtk.GestureClick();
-        contextClick.set_button(3);
+        const contextClick = new Gtk.GestureClick({button: 3});
         contextClick.set_propagation_phase(Gtk.PropagationPhase.CAPTURE);
         this._widgetContainer.add_controller(contextClick);
 
@@ -2406,14 +2504,21 @@ const WidgetGrid = class extends ControlGrid {
         this.setWidgetContainerOnTop(!this._widgetContainerOnTop);
     }
 
-    resizeGrid() {
-        super.resizeGrid();
-
+    resizeWindow() {
+        super.resizeWindow();
         this._widgetContainer.set_size_request(
             this._width,
             this._height
         );
+        this._sizeContainer(this._widgetContainer);
+    }
 
+    resizeGrid() {
+        super.resizeGrid();
+        this._widgetContainer.set_size_request(
+            this._width,
+            this._height
+        );
         this._sizeContainer(this._widgetContainer);
     }
 
